/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : delete.c                                                 */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Contains functions for deleting postit nodes             */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Apr 2002                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <postit.h>      /* Specific header for this dll                       */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

/******************************************************************************/
/* evaluatePostitDelete                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Deletes postits that match in name, scope and CopyDataFrom location.       */
/*                                                                            */
/******************************************************************************/

int evaluatePostitDelete(IN context_t* contextP,
 			   		    IN CciMessage *destinationList,
		                IN CciMessage *exceptionList,
		                IN CciMessage *message)
{
	deployedPostitDelete_t *deployedPostitDeleteP;
	deployedPostitCreate_t *currentPostitCreateP;
	int retCode,
		numMatches;
	postitCreateInstance_t *postitInstanceP,
		                   *nextInstanceP;
    static char *functionName="evaluatePostitDelete";

    TRACE_ENTRY("evaluatePostitDelete", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	/* Using the match conditions specified on the PostitDelete node, look for any
	 * matching postits.  Delete all postits which match on postit name, scope and 
	 * CopyDataFrom values.
	 */

	deployedPostitDeleteP = (deployedPostitDelete_t *)(contextP->nodeDataP);
	TRACE_MID(TRACE_INFO, ("PostitDelete values are"));
	TRACE_MID(TRACE_INFO, ("    PostitName=%S", deployedPostitDeleteP->postitName));
	TRACE_MID(TRACE_INFO, ("    CopyDataFrom=%S", deployedPostitDeleteP->copyFromLocation));
	TRACE_MID(TRACE_INFO, ("    PostitScope=%s", 
		deployedPostitDeleteP->scope==SCOPE_EXECUTION_GROUP ? "ExecutionGroup" :
		deployedPostitDeleteP->scope==SCOPE_BROKER ? "Broker" : "Platform"));


	/* lock the postit chain and look for matching instances */
	TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
	threadMutexLock(postitChainMutex);

	numMatches = 0;

	/* search deployed postit chain looking for matches */
	currentPostitCreateP = postitCreateAnchorP->nextP;

	while (currentPostitCreateP) {
		TRACE_MID(TRACE_INFO, ("Processing possible match at %.8x", currentPostitCreateP));

		/* a match exists if postit name, scope and CopyDataFrom values are the same */
		if (UCcmp(deployedPostitDeleteP->postitName, currentPostitCreateP->postitName) ||
		    UCcmp(deployedPostitDeleteP->copyFromLocation, currentPostitCreateP->copyFromLocation) ||
			deployedPostitDeleteP->scope!=currentPostitCreateP->scope) {
			continue;
		}

		TRACE_MID(TRACE_INFO, ("Matching deployed postit found at %.8x", currentPostitCreateP));
		numMatches++;

		/* remove all postit create instances chained off this deployed postit */
        postitInstanceP = currentPostitCreateP->instanceAnchor.nextP;

		while (postitInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing postit instance at %.8x", postitInstanceP));
			nextInstanceP = postitInstanceP->nextP;

			if (postitInstanceP->global)
			{
				TRACE_MID(TRACE_INFO, ("Shared memory lock required"));

				/* shared memory must be locked around the free operation */
				shmLock(currentPostitCreateP->scope==SCOPE_BROKER);
				freePostit(postitInstanceP);
				shmLock(currentPostitCreateP->scope==SCOPE_BROKER);
			}
			else
			{
				freePostit(postitInstanceP);
			}

			postitInstanceP = nextInstanceP;
		}

		/* process the next in the chain */
		currentPostitCreateP = currentPostitCreateP->nextP;

	}

	TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
	threadMutexUnlock(postitChainMutex);

	if (!numMatches)
	{
		TRACE_MID(TRACE_INFO, ("No matching postit instances found"));

		/* propagate the message to the noMatch terminal */
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'noMatch' "
							   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_NO_MATCH,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'noMatch' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_019, TT_SUCCESS);
	}

	/* propagate message */
	TRACE_MID(TRACE_INFO, ("Propagating message to 'out' terminal"));
	retCode = propagateMessage(contextP,
							   C.TERM_OUT,
							   message,
							   destinationList,
							   exceptionList,
						       FALSE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		RETHROW_IF_EXCEPTION(retCode);
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	TRACE_RETURN(RSN_019, TT_SUCCESS);
}
