/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : request.c                                                */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Contains functions for dealing with request messages     */
/*                   belonging to an aggregate.                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Feb 2000                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <aggregator.h>      /* Specific header for this dll                       */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */
#include <log.h>

/******************************************************************************/
/* Node Utility Functions                                                     */
/*                                                                            */
/* insertAttributeTableEntry                                                  */
/* insertInputTerminalListEntry                                               */
/* insertOutputTerminalListEntry                                              */
/* getInputTerminalHandle                                                     */
/* getOutputTerminalHandle                                                    */
/* traverseMessage                                                                           */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* evaluateAggregateCreate                                                    */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Adds an aggregate member to a new or existing aggregate descriptor.        */
/* The aggregate descriptor is placed in the alarm chain in order of expiry   */
/* time.                                                                      */
/*                                                                            */
/******************************************************************************/
int evaluateAggregateCreate(context_t* contextP,
 			   		        CciMessage *destinationList,
		                    CciMessage *exceptionList,
		                    CciMessage *message)
{
    int retCode,
		sequenceNumber;
	int rc;
	static char *functionName = "evaluateAggregateCreate";
	char msgBuffer[1000],
		 mbString[MAX_STRING_SIZE + 1],
 		 hexBuffer[2*sizeof(MQ_CORREL_ID_LENGTH) + 1];
	deployedAggregate_t *depAggregateP;
	void *correlIdElementP,
		 *messageIdElementP;
    struct CciByteArray byteArrayDetails;
	char byteArray[1000],
		 messageIdArray[MQ_MSG_ID_LENGTH],
		 newMessageIdArray[MQ_MSG_ID_LENGTH];
    CciSize numBytes;
    aggregateInstance_t *currentAggregateP,
		                  *aggregateAnchorP;
	aggregateMember_t *memberP;
	void *newMsg;
	msgIdIndexElement_t *indexP;
	bool_t newAggregate;

    TRACE_ENTRY("evaluateAggregateCreate", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	/* 
	 * Check that the aggregate consists of at least one AggregateCreate node
	 * and at most, one AggregateReply node
	 */
	if (!contextP->nodeDataP || 
		!((deployedAggregate_t *)(contextP->nodeDataP))->aggregateCreateAndReplyExist)
	{
		if (!contextP->nodeDataP)
		{
	        TRACE_MID(TRACE_ERROR, ("Aggregate has not been correctly configured"));

			/* throw an exception */
			sprintf(msgBuffer, "Aggregate has not been correctly configured");

			cciThrowException(&rc,
							  CCI_FATAL_EXCEPTION,
							  __FILE__,
							  __LINE__,
							  functionName,
							  PLUGIN_MESSAGE_CATALOGUE,
							  MJB_INVALID_AGGREGATE_CONFIGURATION, /* messageNumber */ 
							  msgBuffer,           /* traceText */
							  (char *)0);
		}
		else
		{
			depAggregateP = (deployedAggregate_t *)(contextP->nodeDataP);
	        sprintf(msgBuffer, "Aggregate %S is missing either an "
				               "SpAggregateCreate node or an SpAggregateReply "
							   "node", depAggregateP->name);
	        TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

			/* throw an exception */
		    UCToStr(depAggregateP->name,
				    sizeof(mbString),
				    mbString,
				    NULL);
			cciThrowException(&rc,
							  CCI_FATAL_EXCEPTION,
							  __FILE__,
							  __LINE__,
							  functionName,
							  PLUGIN_MESSAGE_CATALOGUE,
							  MJB_MISSING_AGGREGATE_NODE, /* messageNumber */ 
							  msgBuffer,           /* traceText */
#if defined(_AIX)
						      brokerName,
#endif
							  mbString,
							  (char *)0);
		}
	}

	/* extract the message id for use later */
	messageIdElementP = getMessageElement(message, MESSAGE_ID);

	if (!messageIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Message Id not found in message"));
		TRACE_RETURN(RSN_001, TT_MESSAGE_ID_NOT_FOUND);
	}

	/* copy message id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)messageIdArray;
	byteArrayDetails.size = sizeof(messageIdArray);
	numBytes = cniElementByteArrayValue(&rc,
										messageIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Message Id from message"));
		TRACE_RETURN(RSN_002, TT_MESSAGE_ID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Message Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, byteArrayDetails.pointer, MQ_MSG_ID_LENGTH);

	/* extract the correlId and see if we already know about this aggregate instance */
	correlIdElementP = getMessageElement(message, CORREL_ID);

	if (!correlIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Correlation Id not found in message"));
		TRACE_RETURN(RSN_003, TT_CORRELID_NOT_FOUND);
	}

	/* copy correl id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)byteArray;
	byteArrayDetails.size = sizeof(byteArray);
	numBytes = cniElementByteArrayValue(&rc,
										correlIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Correl Id from message"));
		TRACE_RETURN(RSN_004, TT_CORRELID_NOT_FOUND);
	}

	if (traceEnabled)
	{
		char hexBuffer[2*MQ_CORREL_ID_LENGTH + 1];
		dispHex((char *)(byteArrayDetails.pointer),
				MQ_CORREL_ID_LENGTH,
				hexBuffer);
		TRACE_MID(TRACE_INFO, ("CorrelId value is %s", hexBuffer));
	}

	/* ignore messages with correl id's which have not been set */
	if (!memcmp(byteArrayDetails.pointer, MQCI_NONE, MQ_CORREL_ID_LENGTH))
	{
		/* not interested in this message, just propagate it */
		TRACE_MID(TRACE_INFO, ("Message correl id is NULL"));
		TRACE_MID(TRACE_INFO, ("Propagating message to 'out' terminal"));
		retCode = propagateMessage(contextP,
								   L"out",
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
			TRACE_RETURN(RSN_005, TT_PROPAGATION_ERROR);
		}
		TRACE_RETURN(RSN_006, TT_SUCCESS);
	}

	/* search the aggregate instances chain for a match on correl id */
	TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
	threadMutexLock(aggregateChainMutex);

	/* address the aggregate instance chain */
	currentAggregateP = findAggregateInstanceFromContext(contextP,
		                                                 byteArrayDetails.pointer);

	/* if not found, create a new aggregate instance */
	if (!currentAggregateP)
	{
		currentAggregateP = (aggregateInstance_t *)malloc(sizeof(aggregateInstance_t));
		memset(currentAggregateP, 0, sizeof(aggregateInstance_t));

	    TRACE_MID(TRACE_INFO, ("Created new aggregate instance at %.8x",
			                    currentAggregateP));
	    TRACE_MID(TRACE_INFO, ("STORAGE(1): alloc %.8x", currentAggregateP));

		/* initialise the aggregate id */
		memcpy(currentAggregateP->aggregateId, byteArrayDetails.pointer, 
			   MQ_CORREL_ID_LENGTH);

		/* initialise pointer back to deployed aggregate */
		currentAggregateP->deployedAggregateP = 
			                       (deployedAggregate_t *)(contextP->nodeDataP);

		/* initialise the common message id */
		memcpy(currentAggregateP->initialMsgId, messageIdArray, MQ_MSG_ID_LENGTH);
		currentAggregateP->commonMsgId = TRUE;

		/* add to the aggregate instance chain */
		aggregateAnchorP = &(currentAggregateP->deployedAggregateP->aggregateAnchor);

		INSERT_AFTER_IN_LINKED_LIST(aggregateAnchorP, currentAggregateP);

		newAggregate = TRUE;
	}
	else
	{
		newAggregate = FALSE;
		/* do all request messages share a common msg id? */
		if (currentAggregateP->commonMsgId)
		{
			if (memcmp(currentAggregateP->initialMsgId, messageIdArray, MQ_MSG_ID_LENGTH))
			{
				/* 
				 * We only support request messages which share the same MsgId, 
				 * otherwise, how would we know which to use for the CorrelId of the
				 * aggregated reply message?
				 */
				depAggregateP = (deployedAggregate_t *)(contextP->nodeDataP);
				sprintf(msgBuffer, "Request messages with different MsgId values were "
					               "processed by SpAggregateCreate nodes in aggregate %S",
								   depAggregateP->name);

				TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

				/* throw an exception */
				cciThrowException(&rc,
								  CCI_FATAL_EXCEPTION,
								  __FILE__,
								  __LINE__,
								  functionName,
								  PLUGIN_MESSAGE_CATALOGUE,
								  MJB_REQUEST_MSGIDS_NOT_COMMON, /* messageNumber */ 
								  msgBuffer,           /* traceText */
#if defined(_AIX)
								  brokerName,
#endif
								  UCToStr(depAggregateP->name,
										  sizeof(mbString),
										  mbString,
										  NULL),
								  (char *)0);
			}
		}
	}

    TRACE_MID(TRACE_INFO, ("Aggregate messages have common msgId is %s",
 		                    currentAggregateP->commonMsgId ? "TRUE" : "FALSE"));

	/* at this point we are addressing the aggregate descriptor */

	/* create a new aggregate member */
	memberP = (aggregateMember_t *)malloc(sizeof(aggregateMember_t));
	memset(memberP, 0, sizeof(aggregateMember_t));
    TRACE_MID(TRACE_INFO, ("New aggregate member created at %.8x",
		                   memberP));
    TRACE_MID(TRACE_INFO, ("STORAGE(2): alloc %.8x", memberP));

	currentAggregateP->memberCount++;
	currentAggregateP->outstandingReplies++;

    TRACE_MID(TRACE_INFO, ("Aggregate descriptor member count incremented to %d",
	                       currentAggregateP->memberCount));
    TRACE_MID(TRACE_INFO, ("Aggregate descriptor outstanding replies incremented to %d",
	                       currentAggregateP->outstandingReplies));

	/* get sequence number, if any */
	if (getIntegerAttribute(contextP, 
					        ATTR_SEQUENCE_NUMBER,
					        &sequenceNumber) == TT_SUCCESS)
	{
		memberP->sequenceNumber = sequenceNumber;
		TRACE_MID(TRACE_INFO, ("Sequence number set to %d", sequenceNumber));
	}

	/* store address of parent aggregate descriptor in member */
	memberP->aggregateInstanceP = currentAggregateP;
	TRACE_MID(TRACE_INFO, ("Parent aggregate descriptor set to %.8x",
		                   memberP->aggregateInstanceP));

	/* build a new message */
	newMsg = createNewMessage(message);

	if (!newMsg)
	{
		/* error, new message not created */
        TRACE_MID(TRACE_ERROR, ("Failed to create new message"));
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);
		TRACE_RETURN(RSN_007, TT_FAILED_TO_COPY_MESSAGE);
	}

	/* create a new unique message id based on the original one */
	memcpy(newMessageIdArray, messageIdArray, MQ_MSG_ID_LENGTH);
	retCode = genMessageId(newMessageIdArray);
	if (retCode != TT_SUCCESS)
	{
		/* failed to gen a new message id */
        TRACE_MID(TRACE_ERROR, ("Failed to generate a new id for message"));
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);
		TRACE_RETURN(RSN_008, TT_FAILED_TO_CREATE_NEW_MESSAGE_ID);
	}

	/* store old message id in case we want to use it in the future */
	memcpy(memberP->originalMessageId, messageIdArray, MQ_MSG_ID_LENGTH);

	/* store new message id so we can reconstruct the index during recovery */
	memcpy(memberP->newMessageId, newMessageIdArray, MQ_MSG_ID_LENGTH);

	/* address the new message id element */
	messageIdElementP = getMessageElement(newMsg, MESSAGE_ID);
	if (!messageIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Message Id not found in new message"));
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);
		TRACE_RETURN(RSN_009, TT_MESSAGE_ID_NOT_FOUND);
	}

	byteArrayDetails.pointer = (void *)newMessageIdArray;
	byteArrayDetails.size = sizeof(newMessageIdArray);
	cniSetElementByteArrayValue(&rc,
								messageIdElementP,
								&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to set Message Id in message"));
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);
		TRACE_RETURN(RSN_010, TT_FAILED_TO_SET_ID_IN_MESSAGE);
	}

	TRACE_MID(TRACE_INFO, ("Outgoing Message Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, byteArrayDetails.pointer, MQ_MSG_ID_LENGTH);

	/* replace the reply to details, saving the old */
	retCode = swapMessageReplyDetails(newMsg, 
	                                  currentAggregateP->deployedAggregateP->replyToQueueName,
							          queueManagerNameUC,
								      memberP->replyToQueue,
								      sizeof(memberP->replyToQueue),
								      memberP->replyToQueueManager,
								      sizeof(memberP->replyToQueueManager));

	if (retCode != TT_SUCCESS)
	{
		/* failed to modify message reply details */
        TRACE_MID(TRACE_ERROR, ("Failed to replace reply details in message"));
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);
		TRACE_RETURN(RSN_011, TT_FAILED_TO_SWAP_MESSAGE_REPLY_DETAILS);
	}

	TRACE_MID(TRACE_INFO, ("memberP is %.8x", memberP));
	TRACE_MID(TRACE_INFO, ("Saved queueName is '%S'", memberP->replyToQueue));
	TRACE_MID(TRACE_INFO, ("Saved queueManagerName is '%S'", memberP->replyToQueueManager));

	/* add the new member to the aggregate instance, in sequence number order */
	addMemberToAggregateInstance(memberP,
		                         currentAggregateP);

	/* add message to message id index */
	indexP = updateMsgIdIndex(newMessageIdArray);
	indexP->memberP = memberP;
	memberP->indexP = indexP;

	/* reset the aggregate reply timeout */
    setAggregateReplyAlarm(currentAggregateP,
		                   memberP,
						   FALSE);
	
	/* log change to (and creation of) aggregate if recovery is on */
	if (currentAggregateP->deployedAggregateP->persistent)
	{
		/* write a 'new instance' log record if this is a new aggregate */
		if (newAggregate)
		{
			retCode = logNewInstance(currentAggregateP);

			if (retCode != TT_SUCCESS)
			{
				/* raise an exception */
				TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
				threadMutexUnlock(aggregateChainMutex);

				dispHex(currentAggregateP->aggregateId, 
						sizeof(currentAggregateP->aggregateId),
						hexBuffer);

				sprintf(msgBuffer, "Failed to log new aggregate instance record for aggregate %S "
								   "(instance=0x%s)",
									currentAggregateP->deployedAggregateP->name,
									hexBuffer);
				TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

				/* throw an exception */
				cciThrowException(0,
								  CCI_FATAL_EXCEPTION,
								  __FILE__,
								  __LINE__,
								  functionName,
								  PLUGIN_MESSAGE_CATALOGUE,
								  MJB_NEW_AGGREGATE_LOG_FAILURE, /* messageNumber */ 
								  msgBuffer,           /* traceText */
#if defined(_AIX)
								  brokerName,
#endif
								  /* convert aggregate name to multi byte string */
								  UCToStr(currentAggregateP->deployedAggregateP->name,
										  sizeof(mbString),
										  mbString,
										  NULL),
								  hexBuffer,
								  (char *)0);
			}
		}

		retCode = logMember(memberP);

		if (retCode != TT_SUCCESS)
		{
			/* raise an exception */
			TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
			threadMutexUnlock(aggregateChainMutex);

			dispHex(currentAggregateP->aggregateId, 
				    sizeof(currentAggregateP->aggregateId),
					hexBuffer);

			sprintf(msgBuffer, "Failed to log aggregate member details for aggregate %S "
				               "(instance=0x%s)",
                                currentAggregateP->deployedAggregateP->name,
								hexBuffer);
			TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

			/* throw an exception */
			cciThrowException(0,
							  CCI_FATAL_EXCEPTION,
							  __FILE__,
							  __LINE__,
							  functionName,
							  PLUGIN_MESSAGE_CATALOGUE,
							  MJB_AGGREGATE_UPDATE_LOG_FAILURE, /* messageNumber */ 
							  msgBuffer,           /* traceText */
#if defined(_AIX)
							  brokerName,
#endif
							  /* convert aggregate name to multi byte string */
							  UCToStr(currentAggregateP->deployedAggregateP->name,
									  sizeof(mbString),
									  mbString,
									  NULL),
							  hexBuffer,
							  (char *)0);
		}
	}

	/* signal the alarm thread to process new aggregate descriptor */
	threadCondSignal(alarmThreadWorkCondVar);

	TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
	threadMutexUnlock(aggregateChainMutex);

    /* propagate the new message to the output terminal */
	TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x", newMsg));
	retCode = propagateMessage(contextP,
							   L"out",
							   newMsg,
							   destinationList,
							   exceptionList,
							   TRUE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	/* delete the new message */
    cniDeleteMessage(&rc,
		             newMsg);

	TRACE_RETURN(RSN_013, TT_SUCCESS);
}

/******************************************************************************/
/* addMemberToAggregateInstance                                               */
/******************************************************************************/
void addMemberToAggregateInstance(IN aggregateMember_t *memberP,
		                          IN aggregateInstance_t *aggregateP)
{
	int thisSequenceNumber;
	aggregateMember_t *currentMemberP;
    TRACE_ENTRY("addMemberToAggregateInstance", ("memberP=%.8x, aggregateP=%.8x",
		                                        memberP, aggregateP));

	/* add the member to the aggregate, in sequence number order */
	thisSequenceNumber = memberP->sequenceNumber;
	currentMemberP = &(aggregateP	->memberAnchor);

	while(TRUE)
	{
		if (!currentMemberP->nextP ||
			currentMemberP->nextP->sequenceNumber > thisSequenceNumber)
		{
			break;
		}

		currentMemberP = currentMemberP->nextP;
	}

    /* add member after current member */
	INSERT_AFTER_IN_LINKED_LIST(currentMemberP, memberP);

	TRACE_RETURN_VOID(RSN_014);
}

/******************************************************************************/
/* replaceMessageCorrelId                                                     */
/******************************************************************************/
void *replaceMessageCorrelId(IN void *message, 
							 IN char *correlId)
{
	void *newMsg,
		 *correlIdElementP;
	int rc;
    struct CciByteArray byteArrayDetails;

    TRACE_ENTRY("replaceMessageCorrelId", ("message=%.8x, correlId=%.8x",
		                                    message, correlId));

	/* Create a new message */
    newMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message));
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message (cniCreateMessage FAILED)"));
		TRACE_RETURN_TYPE(RSN_015, void *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

	/* copy old message to new */
	rc = copyMessage(newMsg, message);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new"));
		TRACE_RETURN_TYPE(RSN_016, void *, NULL);
	}
	TRACE_MID(TRACE_INFO, ("Original message copied to new OK"));

	/* locate the correl id element in new message structure */
	correlIdElementP = getMessageElement(newMsg, CORREL_ID);

	if (!correlIdElementP)
	{
		/* generate an error, exit */
        TRACE_MID(TRACE_ERROR, ("Correl Id element not found in new message"));
		TRACE_RETURN_TYPE(RSN_017, void *, NULL);
	}

	/* replace the correl id with the supplied value */
	byteArrayDetails.pointer = (void *)correlId;
	byteArrayDetails.size = MQ_MSG_ID_LENGTH;
	TRACE_MID(TRACE_INFO, ("Setting correl id to..."));
	TRACE_MIDHEX(TRACE_INFO, byteArrayDetails.pointer, MQ_MSG_ID_LENGTH);

	cniSetElementByteArrayValue(&rc,
								correlIdElementP,
								&byteArrayDetails);

	if (rc != CCI_SUCCESS)
	{
	    TRACE_MID(TRACE_ERROR, ("Failed to set correl id (cniSetElementByteArrayValue rc=%d)", rc));
		TRACE_RETURN_TYPE(RSN_018, void *, NULL);
	}
    TRACE_MID(TRACE_INFO, ("Correl id successfully set"));
	
	/* return the new message to caller */
	TRACE_RETURN_TYPE(RSN_019, void *, newMsg);
}

/******************************************************************************/
/* genMessageId                                                               */
/******************************************************************************/
#include <uuid.h>
int genMessageId(OUT char *msgId)
{
	int rc;
	UUID uuid;
	int offset;

    TRACE_ENTRY("genMessageId", ("msgId=%.8x", msgId));

	/* 
	 * generate a uuid as basis for message id but single thread since not all
	 * internal functions (networking functions) are thread safe
	 */
	threadMutexLock(uuidGenMutex);
	rc = uuid_create(&uuid);
	threadMutexUnlock(uuidGenMutex);

	if (rc != UUID_SUCCESS)
	{
		if (rc == UUID_FAILED_TO_GET_MACHINE_HARDWARE_ADDRESS)
		{
   	        TRACE_RETURN(RSN_020, TT_FAILED_TO_GET_MACHINE_HARDWARE_ADDRESS);
		}
        TRACE_RETURN(RSN_021, UUID_FAILED);
	}

	/* copy new value into last 16 bytes of callers message id storage */
	offset = MQ_MSG_ID_LENGTH-sizeof(uuid);
	memcpy(msgId+offset, &uuid, sizeof(uuid));

	TRACE_RETURN(RSN_022, TT_SUCCESS);
}

/******************************************************************************/
/* findAggregateInstanceFromContext                                           */
/******************************************************************************/
aggregateInstance_t *findAggregateInstanceFromContext(IN context_t *contextP,
														IN char *aggregateId)
{
	aggregateInstance_t *aggregateP;

    TRACE_ENTRY("findAggregateInstanceFromContext", ("contextP=%.8x", contextP));

	/* address the first aggregate in the list */
    aggregateP = ((deployedAggregate_t *)(contextP->nodeDataP))->aggregateAnchor.nextP;

	while(aggregateP)
	{
		if (traceEnabled)
		{
			char hexBuffer[2*sizeof(aggregateP->aggregateId) + 1];
			dispHex((char *)(aggregateP->aggregateId),
				    sizeof(aggregateP->aggregateId),
					hexBuffer);
		    TRACE_MID(TRACE_INFO, ("Processing aggregate at %.8x (%s)", 
				                   aggregateP, hexBuffer));
		}
		else
		{
		    TRACE_MID(TRACE_INFO, ("Processing aggregate at %.8x", aggregateP));
		}

#if 0
		TRACE_MID(TRACE_INFO, ("Aggregate id value is..."));
		TRACE_MIDHEX(TRACE_INFO, aggregateP->aggregateId, MQ_CORREL_ID_LENGTH);
#endif

		/* check for matching aggregate id */
		if (memcmp(aggregateId, aggregateP->aggregateId, MQ_CORREL_ID_LENGTH))
		{
			aggregateP = aggregateP->nextP;
			continue;
		}

	    TRACE_MID(TRACE_INFO, ("Matching aggregate id found"));
		break;
	}

	TRACE_RETURN_TYPE(RSN_023, aggregateInstance_t *, aggregateP);
}


